//------------------------------------------------------------------------------------------
// Intelligent Global Path Planner with Replanning
//------------------------------------------------------------------------------------------

unit GraphIGPPR;

interface

uses
  Graph, Simulator, CollisionDetection, Polygon, Classes, Math, Distance,
  Dialogs, SysUtils;

type
  TCell = record            // bunka
    evaluation: integer;    // ohodnoceni bunky
    i,j: integer;           // pozice bunky v poli
  end;

  TTree = array of TVert;
  TFD = array of array of TCell;

  TPosition = class         // pozice bunky v poli
    i,j: integer;
  end;


  TGraphIGPPR = class (TGraph)
  private
    fMaxX: integer;   // celkova sirka plochy
    fMaxY: integer;   // celkova vyska plochy
    fEndD: double;    // max.odchylka od cilove konf.(podm. ukonceni vypoctu)
    fFruit: vector_q; // ukazatel na final node
      // parametry modelu auta
    fCarL: double;		   // vzdalenost os
	  fMaxFi: double;		   // maximalni natoceni prednich kol
	  fCarWidth: double;	 // sirka auta
  	fCarOver: double;		 // presah pres osu vzadu a vpredu auta
	  fdt: integer;		  	 // velikost bunky
    fdti: integer;       // integracni krok
    fStop: boolean;      // zastaveni vypoctu
    fCostMap: integer;   // typ mapy ohodnoceni (4 nebo 8 smeru)
    ColX,ColY: integer;  // parametr pro navyseni prekazek
    fDrawTree: boolean;  // vykreslovani stromu behem vypoctu
    fDrawCells: boolean; // vykreslovani bunek
    Simulator: TSimulator;
    Distance: TDistance;
    CollisionDetector: TCollisionDetection;
    ObsList: TList;

    FD: TFD;             // 2n-pole predstavujici rozdeleny prostoru do mrizky
    SolVertices: TTree;  // pole vertex na cest k cli
    function Evaluate(aNode:TNode):TEvaluation; //metoa pro ohodnocen cesty
    function IsOccupied(aCell:TCell): boolean;
    function PointLocation(aConf:vector_q):TPosition;
  protected
    function RulesCount:integer;override;

    function IsEqualData(aNode1,aNode2:TNode):boolean;override;
    procedure DrawTreeProc(a,b: TVert);virtual;abstract;
    procedure DrawCellsProc(a,b: vertex);virtual;abstract;
  public
    constructor Create(aWidth,aLength:integer;aFruit:vector_q);
    destructor Destroy;override;
    procedure SetObstacles(aObstacles: TList);
    procedure DrawTree(q_near,q_new: vector_q);
    procedure DrawCells;
    procedure CreateCostMap;
    procedure FindPath;
    procedure ResetSol;
    function IGPPR(aRoot,aFruit:vector_q;aMaxDepth:integer):boolean;
    function ActionInput(expandingNode,aFruit,aRoot:vector_q;action:integer):vector_u;
    property xSimulator:TSimulator read Simulator write Simulator;
    property bDrawCells:boolean read fDrawCells write fDrawCells;
    property bDrawTree:boolean read fDrawTree write fDrawTree;
    property MaxFi:double read fMaxFi write fMaxFi;
    property EndD:double read fEndD write fEndD;
    property SolTree:TTree read SolVertices;
    property MaxX:integer read fMaxX;
    property MaxY:integer read fMaxY;
    property Stop:boolean read fStop write fStop;
    property CostMap:integer read fCostMap write fCostMap;
    property dt:integer read fdt write fdt;
    property dti:integer read fdti write fdti;
    property D:TFD read FD;
  end;

implementation

{ TIGPPR }

constructor TGraphIGPPR.Create(aWidth,aLength:integer;aFruit:vector_q);
begin
  inherited create;
  fMaxX:=aWidth;
  fMaxY:=aLength;
  fdt:=7;     // velikost bunky
  fdti:=8;    // integracni krok
  fEndD:=8;
  fStop:=false;
  fCostMap:=4;
  fCarL:=15;
  fCarOver:=6;
  fCarWidth:=15;
  fMaxFi:=35*(Pi/180);
  ColX:=9;
  ColY:=9;
  fFruit:=aFruit;
  fDrawTree:=true;
  fDrawCells:=false;
  Simulator:=TSimulator.Create;
  Simulator.SetLength(fCarL);
  Distance:=TDistance.Create;
  Distance.SetA(fCarL*fCarL);
  CollisionDetector:=TCollisionDetection.Create;
end;

destructor TGraphIGPPR.Destroy;
begin
  Setlength(FD,0,0);
  Distance.Free;
  Simulator.Free;
  CollisionDetector.Free;
  inherited;
end;

function TGraphIGPPR.RulesCount: integer;
begin
  result:=6;
end;

procedure TGraphIGPPR.SetObstacles(aObstacles: TList);
begin
  ObsList:=aObstacles;
end;

  // vraci true pokud je bunka obsazena prekazkou
function TGraphIGPPR.IsOccupied(aCell:TCell): boolean;
var x,y,i,j: integer;
  v: TVertex;
  pol: TPolygon;
  bool: boolean;
begin
  result:=false;
  pol:=TPolygon.Create;
    // pozice leveho horniho vrcholu bunky
  y:=fdt*aCell.i;
  x:=fdt*aCell.j;
  v:=TVertex.Create;
  v.x:=x-ColX; //zvetseni o parametr ColX,ColY
  v.y:=y-ColY;
  pol.AddVertex(v);
    // pozice praveho horniho vrcholu bunky
  y:=fdt*aCell.i;
  x:=fdt*aCell.j+fdt;
  v:=TVertex.Create;
  v.x:=x+ColX;
  v.y:=y-ColY;
  pol.AddVertex(v);
    // pozice praveho dolniho vrcholu bunky
  y:=fdt*aCell.i+fdt;
  x:=fdt*aCell.j+fdt;
  v:=TVertex.Create;
  v.x:=x+ColX;
  v.y:=y+ColY;
  pol.AddVertex(v);
    // pozice leveho dolniho vrcholu bunky
  y:=fdt*aCell.i+fdt;
  x:=fdt*aCell.j;
  v:=TVertex.Create;
  v.x:=x-ColX;
  v.y:=y+ColY;
  pol.AddVertex(v);
  for i:=0 to ObsList.Count-1 do
    begin
      bool:=CollisionDetector.CollisionDetection(pol,TPolygon(ObsList.Items[i]));
      if bool then
        begin
          result:=true;
          exit;
        end;
    end;
end;

  // vytvoreni mapy heurist.cen
procedure TGraphIGPPR.CreateCostMap;
var i,j: integer;
    w,l: integer;
    cell,cell2: TCell;
    cellInit: TCell;
    exp,posFruit,pos: TPosition;
    open: TList;
begin
  w:=trunc(fMaxX/fdt);
  l:=trunc(fMaxY/fdt);

  try
    SetLength(FD,0,0);
    SetLength(FD,l,w);
  except
    raise Exception.Create('Creating CostMAp Error!');
  end;
  cellInit.evaluation:=-1;
  for i:=0 to l-1 do
    for j:=0 to w-1 do
      begin
        cellInit.i:=i;
        cellInit.j:=j;
        FD[i,j]:=cellInit;
      end;
    // zjisti bunku Rootu
  posFruit:=PointLocation(fFruit);
  FD[posFruit.i,posFruit.j].evaluation:=0;

  open:=TList.create;
  open.Add(posFruit);

 while open.Count>0 do
 begin
  exp:=open.Items[0];
  open.Delete(0);
  cell2:=FD[exp.i,exp.j];
    // 8.smerova metoda
  for i:=1 to fCostMap do
    begin
      case i of
      1: begin //doleva
            if (exp.j=0) then continue;
            cell:=FD[exp.i,exp.j-1];
            if (exp.i=posFruit.i) and (exp.j-1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+17;
                  FD[exp.i,exp.j-1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i; pos.j:=exp.j-1;
                  open.Add(pos);
              end;
         end;
      2: begin //nahoru
            if (exp.i=0) then continue;
            cell:=FD[exp.i-1,exp.j];
            if  (exp.i-1=posFruit.i) and (exp.j=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+17;
                  FD[exp.i-1,exp.j]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i-1; pos.j:=exp.j;
                  open.Add(pos);
              end;
         end;
      3: begin //doprava
            if (exp.j=w-1) then continue;
            cell:=FD[exp.i,exp.j+1];
            if (exp.i=posFruit.i) and (exp.j+1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+17;
                  FD[exp.i,exp.j+1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i; pos.j:=exp.j+1;
                  open.Add(pos);
              end;
         end;
      4: begin //dolu
            if (exp.i=l-1) then continue;
            cell:=FD[exp.i+1,exp.j];
            if (exp.i+1=posFruit.i) and (exp.j=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+17;
                  FD[exp.i+1,exp.j]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i+1; pos.j:=exp.j;
                  open.Add(pos);
              end;
         end;
         // diagonalne
      5: begin //vlevo-dolu
            if (exp.i=l-1)or(exp.j=0) then continue;
            cell:=FD[exp.i+1,exp.j-1];
            if (exp.i+1=posFruit.i) and (exp.j-1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+24;
                  FD[exp.i+1,exp.j-1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i+1; pos.j:=exp.j-1;
                  open.Add(pos);
              end;
         end;
       6: begin //vlevo-nahoru
            if (exp.i=0)or(exp.j=0) then continue;
            cell:=FD[exp.i-1,exp.j-1];
            if (exp.i-1=posFruit.i) and (exp.j-1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+24;
                  FD[exp.i-1,exp.j-1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i-1; pos.j:=exp.j-1;
                  open.Add(pos);
              end;
         end;
       7: begin //vpravo-nahoru
            if (exp.i=0)or(exp.j=w-1) then continue;
            cell:=FD[exp.i-1,exp.j+1];
            if (exp.i-1=posFruit.i) and (exp.j+1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+24;
                  FD[exp.i-1,exp.j+1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i-1; pos.j:=exp.j+1;
                  open.Add(pos);
              end;
         end;
       8: begin //vpravo-dolu
            if (exp.i=l-1)or(exp.j=w-1) then continue;
            cell:=FD[exp.i+1,exp.j+1];
            if (exp.i+1=posFruit.i) and (exp.j+1=posFruit.j) then continue;
            if (cell.evaluation=-1) then
              begin
                if IsOccupied(cell) then continue
                else
                  cell.evaluation:=cell2.evaluation+24;
                  FD[exp.i+1,exp.j+1]:=cell;
                  pos:=TPosition.Create;
                  pos.i:=exp.i+1; pos.j:=exp.j+1;
                  open.Add(pos);
              end;
         end;
      end;
    end;
 end;

end;

  // vraci pozici bunky ve ktere by se mela nachazet zadana konfigurace
function TGraphIGPPR.PointLocation(aConf: vector_q): TPosition;
var pos: TPosition;
  i,j: integer;
begin
  i:=trunc(aConf.x/fdt);
  j:=trunc(aConf.y/fdt);
  pos:=TPosition.Create;
  pos.i:=j; pos.j:=i;
  result:=pos;
end;

  // ohodnoceni vrcholu
function TGraphIGPPR.Evaluate(aNode:TNode):TEvaluation;
var i,j,f,g,h:integer;
  pos: TPosition;
begin
     result:=maxint;
     if not(aNode is vector_q) then exit;
     pos:=PointLocation(vector_q(aNode));

     h:=FD[pos.i,pos.j].evaluation; // h: ohodnoceni bunky
     g:=aNode.Depth;                // g: pocet akci(hloubka)
     f:=g+h;                        // hodnotc fce
     aNode.Evaluation := f;
     result:=f;
end;

  // vraci jednu z 6ti moznych vstupnich akci
function TGraphIGPPR.ActionInput(expandingNode,aFruit,aRoot:vector_q; action:integer):vector_u;
var i,j,v: integer;
    fi,d,minD: double;
    q_new,q_opt: vector_q;
    u,u_opt: vector_u;
    pos: TPosition;
    pom: TNode;
begin
  result:=nil;
  case action of
    2: begin
        v:=1; fi:=-fMaxFi;
       end;
    1: begin
        v:=1; fi:=0;
       end;
    3: begin
        v:=1; fi:=fMaxFi;
       end;
    5: begin
        v:=-1; fi:=-fMaxFi;
       end;
    4: begin
        v:=-1; fi:=0;
       end;
    6: begin
        v:=-1; fi:=fMaxFi;
       end;
  end;

  u:=vector_u.create(expandingNode.ID,0,v,fi);
  q_new:=Simulator.Integrate(expandingNode,u,fdti,2);
  if (q_new.x<5) or (q_new.y<5) or (q_new.x>795) or (q_new.y>595) then exit;

  pos:=PointLocation(q_new);
  if TCell(FD[pos.i,pos.j]).evaluation=-1 then
      begin
        u.Free; q_new.Free;
        exit;
      end;
  q_new.Depth:=expandingNode.Depth+1;
  Evaluate(q_new);
  if (expandingNode.Depth>1) then   // vraci jen uzly s lepsim ohodnocenim oproti predchudci
    if (expandingNode.Evaluation+5<q_new.Evaluation) then
      begin
        u.Free; q_new.Free;
        exit;
      end;
  result:=u;
end;

function TGraphIGPPR.IGPPR(aRoot,aFruit:vector_q;aMaxDepth:integer):boolean;
var Open,Closed:TList;
    Found:boolean;
    ExpandingNode,newNode,pom:vector_q;
    u,newEdge: vector_u;
    i,newID,eval,depth:integer;
    pos: TPosition;
    d: double;
begin
     result:=false;
     CleanAll;
     if fDrawCells then DrawCells;
     AddNode(aRoot);
     if IsEqualData(aRoot,aFruit) then
      begin
        Solution.Add(aRoot); result:=true; exit;
      end;
     Open:=TList.Create;
     Closed:=TList.Create;
     ExpandingNode:=nil;
     Found:=false;
     pos:=PointLocation(aRoot);
     aRoot.Evaluation:=TCell(FD[pos.i,pos.j]).evaluation;
     Open.Add(aRoot);
     while (not found) and (Open.count>0) do
       begin
          if (fStop) then begin   // zastaveni vypoctu
            CleanAll;
            exit;
          end;
          ExpandingNode:= vector_q(open.Items[Open.Count-1]);
          Closed.Add(ExpandingNode);
          Open.Delete(Open.Count-1);
          for i:=1 to RulesCount do
            begin     // vraci vstupni akce (pouze uzly s lepsim ohodnocenim)
              u:= ActionInput(expandingNode,aFruit,aRoot,i);
              if (u=nil) then continue;  // kolizni akce
              newNode:=Simulator.Integrate(expandingNode,u,fdti,2);
                newID:=getMaxNodeID+1;
                newNode.ID:=newID;
                newNode.Depth:=expandingNode.Depth+1;
                Evaluate(newNode);
              AddNode(newNode);         // pridani uzlu do stromu
                u.ToID:=newNode.ID;
              AddEdge(u);               // pridani hrany
              if IsEqualData(newNode,aFruit) then
                begin
                  Solution.Add(NewNode);
                  RebuildSolution;
                  found:=true;
                  break;
                end;
              if newNode.Depth < aMaxDepth+1 then
                begin
                  SortedList(open,newNode);
                  if fDrawTree then DrawTree(expandingNode,newNode);
                end
              else begin
                Open.Clear;
                Closed.Clear;
                exit;
              end;
            end;
       end;
     if found then result:=true;
     Open.Free;
     Closed.Free;
end;

procedure TGraphIGPPR.FindPath;
var q_near: vector_q;
         v: TVert;
       i,l: integer;
begin
  if (Solution.Count=0) then exit;
  SetLength(SolVertices,0);
  for i:=0 to Solution.Count-1 do
  begin
    v.x:=trunc(vector_q(Solution.Items[i]).x);
    v.y:=trunc(vector_q(Solution.Items[i]).y);
    SetLength(SolVertices,i+1);
    SolVertices[i]:=v;
  end;
end;

function TGraphIGPPR.IsEqualData(aNode1, aNode2: TNode): boolean;
var d: double;
begin
  result:=false;
  if not(aNode1 is vector_q)or not(aNode2 is vector_q) then exit;
  d:=Distance.Distance(vector_q(aNode1),vector_q(aNode2));
  if (d<fEndD) then result:=true;
end;

procedure TGraphIGPPR.DrawTree(q_near, q_new: vector_q);
var a,b: TVert;
    l: integer;
begin
  a.x:=trunc(q_near.x);
  a.y:=trunc(q_near.y);

  b.x:=trunc(q_new.x);
  b.y:=trunc(q_new.y);
  DrawTreeProc(a,b);
end;

procedure TGraphIGPPR.DrawCells;
var a,b: vertex;
    w,l,i: integer;
begin
  w:=trunc(fMaxX/fdt);
  l:=trunc(fMaxY/fdt);
  for i:=1 to w do
    begin
      a.x:=i*fdt;
      a.y:=0;
      b.x:=i*fdt;
      b.y:=l*fdt;
      DrawCellsProc(a,b);
    end;
  for i:=1 to l do
    begin
      a.x:=0;
      a.y:=i*fdt;
      b.x:=w*fdt;
      b.y:=i*fdt;
      DrawCellsProc(a,b);
    end;
end;


procedure TGraphIGPPR.ResetSol;
begin
  SetLength(SolVertices,0);
end;

end.

