//------------------------------------------------------------------------------------------
// Incorporating State Space Discretization
//------------------------------------------------------------------------------------------

unit GraphISSD;

interface

uses
  Graph, Simulator, Polygon, Distance, CollisionDetection, Dialogs, Classes, Math;

type
  TCell2 = record         // bunka
    visited: boolean;     // navstivena search alg.
    occupied: boolean;    // obsazena prekazkou
    node: vector_q;       // ukazatel na konf.
    i,j: integer;         // pozice bunky v poli
  end;

  TTree = array of TVert;
  TFD = array of array of TCell2;

  TPosition = class         // pozice bunky v poli
    i,j: integer;
  end;

  TGraphISSD = class (TGraph)
  private
    fMaxX: integer;   // celkova sirka plochy
    fMaxY: integer;   // celkova vyska plochy
    fEndD: double;    // max.odchylka od cilove konf.(podm. ukonceni vypoctu)
    fGoal: vector_q;  // ukazatel na goal node
          // parametry modelu auta
    fCarL: double;		  // vzdalenost os
	  fMaxFi: double;		  // maximalni natoceni prednich kol
	  fCarWidth: double;	// sirka auta
  	fCarOver: double;		// presah pres osu vzadu a vpredu auta

    ColX,ColY: integer;  // parametr pro navyseni prekazek
    fdt: integer;		  	 // velikost bunky
    fdti: integer;       // integracni krok
    fStop: boolean;      // zastaveni vypoctu
    FD: TFD;             // 2d-pole predstavujici rozdeleny prostoru do mrizky
    SolVertices: TTree;  // pole vertex na cest k cli
    fDrawTree: boolean;  // vykreslovani stromu behem vypoctu
    fDrawCells: boolean; // vykreslovani bunek

    Simulator: TSimulator;
    Distance: TDistance;
    CollisionDetector: TCollisionDetection;
    ObsList: TList;
    function AddNodeToGraph(expandingNode:vector_q; i:integer):vector_q;
    procedure Evaluate(aNode:TNode);

    function IsOccupied(aCell:TCell2): boolean;
    function PointLocation(aConf:vector_q):TPosition;
  protected
    function RulesCount:integer; override;
    function IsEqualData(aNode1,aNode2:TNode):boolean; override;
    procedure DrawTreeProc(a,b: TVert); virtual; abstract;
    procedure DrawCellsProc(a,b: vertex); virtual; abstract;
  public
    constructor Create(aWidth,aLength:integer; aGoal:vector_q);
    destructor Destroy;override;
    procedure SetObstacles(aObstacles: TList);
    procedure DrawTree(q_near, q_new: vector_q);
    procedure DrawCells;
    procedure FindPath;
    procedure Decomposing;
    procedure ResetSol;
    procedure Default;
    function ISSD(aInit,aGoal:vector_q; aMaxDepth:integer):boolean;
    property xSimulator:TSimulator read Simulator write Simulator;
    property bDrawCells:boolean read fDrawCells write fDrawCells;
    property bDrawTree:boolean read fDrawTree write fDrawTree;
    property MaxFi:double read fMaxFi write fMaxFi;
    property EndD:double read fEndD write fEndD;
    property SolTree:TTree read SolVertices;
    property MaxX:integer read fMaxX;
    property MaxY:integer read fMaxY;
    property Stop:boolean read fStop write fStop;
    property dt:integer read fdt write fdt;
    property dti:integer read fdti write fdti;
    property D:TFD read FD;
  end;

implementation

{ TGraphISSD }

constructor TGraphISSD.Create(aWidth,aLength:integer; aGoal:vector_q);
begin
  inherited create;
  fMaxX:=aWidth;
  fMaxY:=aLength;
  fGoal:=aGoal;
  fdt:=7;     // velikost bunky
  fdti:=8;    // integracni krok  !! dti>=dt !! jinak neopust bunku
  fEndD:=8;
  fStop:=false;
  fCarL:=15;
  fCarOver:=6;
  fCarWidth:=15;
  fMaxFi:=35*(Pi/180);
  ColX:=9;
  ColY:=9;
  fDrawTree:=true;
  fDrawCells:=false;

  Simulator:=TSimulator.Create;
  Simulator.SetLength(fCarL);
  Distance:=TDistance.Create;
  Distance.SetA(fCarL*fCarL);
  CollisionDetector:=TCollisionDetection.Create;
end;

destructor TGraphISSD.Destroy;
begin
  Setlength(FD,0,0);
  Distance.Free;
  Simulator.Free;
  CollisionDetector.Free;
  inherited;
end;

procedure TGraphISSD.SetObstacles(aObstacles: TList);
begin
  ObsList:=aObstacles;
end;

function TGraphISSD.RulesCount: integer;
begin
  result:=6;
end;

  // vraci pozici bunky zadane konfigurace
function TGraphISSD.PointLocation(aConf: vector_q): TPosition;
var pos: TPosition;
  i,j: integer;
begin
  i:=trunc(aConf.x/fdt);
  j:=trunc(aConf.y/fdt);
  pos:=TPosition.Create;
  pos.i:=j; pos.j:=i;
  result:=pos;
end;

  // ohodnoceni vrcholu
procedure TGraphISSD.Evaluate(aNode:TNode);
var i,j,f,g,h:integer;
       pos: TPosition;
begin
     if not(aNode is vector_q) then exit;
     g:=aNode.Depth;
     h:=trunc(Distance.Distance(vector_q(aNode),vector_q(fGoal)));
     f:=g+h;  // pocet kroku(hloubka) + vzdalenost k cily
     aNode.Evaluation := f;
end;

  // vraci true pokud je bunka obsazena prekazkou
function TGraphISSD.IsOccupied(aCell:TCell2): boolean;
var x,y,i,j: integer;
  v: TVertex;
  pol: TPolygon;
  bool: boolean;
begin
  result:=false;
  pol:=TPolygon.Create;

    // pozice leveho horniho vrcholu bunky
  y:=fdt*aCell.i;
  x:=fdt*aCell.j;
  v:=TVertex.Create;
  v.x:=x-ColX; //zvetseni o parametr ColX,ColY
  v.y:=y-ColY;
  pol.AddVertex(v);
    // pozice praveho horniho vrcholu bunky
  y:=fdt*aCell.i;
  x:=fdt*aCell.j+fdt;
  v:=TVertex.Create;
  v.x:=x+ColX;
  v.y:=y-ColY;
  pol.AddVertex(v);
    // pozice praveho dolniho vrcholu bunky
  y:=fdt*aCell.i+fdt;
  x:=fdt*aCell.j+fdt;
  v:=TVertex.Create;
  v.x:=x+ColX;
  v.y:=y+ColY;
  pol.AddVertex(v);
    // pozice leveho dolniho vrcholu bunky
  y:=fdt*aCell.i+fdt;
  x:=fdt*aCell.j;
  v:=TVertex.Create;
  v.x:=x-ColX;
  v.y:=y+ColY;
  pol.AddVertex(v);

  for i:=0 to ObsList.Count-1 do
    begin
      bool:=CollisionDetector.CollisionDetection(pol,TPolygon(ObsList.Items[i]));
      if bool then
        begin
          result:=true;
          exit;
        end;
    end;
end;


  // rozklad stav.prostoru do bunek
procedure TGraphISSD.Decomposing;
var i,j: integer;
    w,l: integer;
    cell,cell2,cellInit: TCell2;
    exp,posInit,pos: TPosition;
    open: TList;
begin
  w:=trunc(fMaxX/fdt);
  l:=trunc(fMaxY/fdt);
  SetLength(FD,l,w);
  cellInit.visited:=false;
  cellInit.occupied:=true;
  cellInit.node:=nil;
  for i:=0 to l-1 do
    for j:=0 to w-1 do
      begin
        cellInit.i:=i;
        cellInit.j:=j;
        FD[i,j]:=cellInit;
      end;
    // zjisti bunku Startu
  posInit:=PointLocation(fGoal);
  FD[posInit.i,posInit.j].node:=fGoal;
  FD[posInit.i,posInit.j].visited:=true;
  FD[posInit.i,posInit.j].occupied:=false;    // ?? hlidat to   kontrolovat aspon CollsiDetect

  open:=TList.create;
  open.Add(posInit);

while open.Count>0 do
 begin
  exp:=open.Items[0];
  open.Delete(0);
  cell2:=FD[exp.i,exp.j];
    // 4.smerova metoda
  for i:=1 to 4 do
    begin
      case i of
      1: begin //doleva
            if (exp.j=0) then continue;
            cell:=FD[exp.i,exp.j-1];
            if (cell.i=posInit.i) and (cell.j=posInit.j) then continue;
            if (cell.occupied=true) then
              begin
                if IsOccupied(cell) then continue
                  else
                    cell.occupied:=false;
                    FD[cell.i,cell.j]:=cell;
                    pos:=TPosition.Create;
                    pos.i:=cell.i; pos.j:=cell.j;
                    open.Add(pos);
              end;
         end;
      2: begin //nahoru
            if (exp.i=0) then continue;
            cell:=FD[exp.i-1,exp.j];
            if  (cell.i=posInit.i) and (cell.j=posInit.j) then continue;
            if (cell.occupied=true) then
              begin
                if IsOccupied(cell) then continue
                  else
                    cell.occupied:=false;
                    FD[cell.i,cell.j]:=cell;
                    pos:=TPosition.Create;
                    pos.i:=cell.i; pos.j:=cell.j;
                    open.Add(pos);
              end;
         end;
      3: begin //doprava
            if (exp.j=w-1) then continue;
            cell:=FD[exp.i,exp.j+1];
            if  (cell.i=posInit.i) and (cell.j=posInit.j) then continue;
            if (cell.occupied=true) then
              begin
                if IsOccupied(cell) then continue
                  else
                    cell.occupied:=false;
                    FD[cell.i,cell.j]:=cell;
                    pos:=TPosition.Create;
                    pos.i:=cell.i; pos.j:=cell.j;
                    open.Add(pos);
              end;
         end;
      4: begin //dolu
            if (exp.i=l-1) then continue;
            cell:=FD[exp.i+1,exp.j];
            if  (cell.i=posInit.i) and (cell.j=posInit.j) then continue;
            if (cell.occupied=true) then
              begin
                if IsOccupied(cell) then continue
                  else
                    cell.occupied:=false;
                    FD[cell.i,cell.j]:=cell;
                    pos:=TPosition.Create;
                    pos.i:=cell.i; pos.j:=cell.j;
                    open.Add(pos);
              end;
         end;
      end;
    end;
 end;

end;


  // defaultni nastaveni mrizky stav.prostoru
procedure TGraphISSD.Default;
var i,j,w,l: integer;
     posInit: TPosition;
begin
  w:=trunc(fMaxX/fdt);
  l:=trunc(fMaxY/fdt);
  for i:=0 to l-1 do
    for j:=0 to w-1 do
      begin
        FD[i,j].visited:=false;
        FD[i,j].node:=nil;
      end;
    // zjisti bunku Startu
  posInit:=PointLocation(fGoal);
  FD[posInit.i,posInit.j].node:=fGoal;
  FD[posInit.i,posInit.j].visited:=true;
  FD[posInit.i,posInit.j].occupied:=false;
end;


  // vraci nove vytvorenou konfiguraci
function TGraphISSD.AddNodeToGraph(expandingNode:vector_q; i:integer):vector_q;
var  v,newID: integer;
    fi: double;
    q_new: vector_q;
    u: vector_u;
    pos: TPosition;
begin
  result:=nil;
  case i of
    2: begin
        v:=1; fi:=-fMaxFi;
       end;
    1: begin
        v:=1; fi:=0;
       end;
    3: begin
        v:=1; fi:=fMaxFi;
       end;
    5: begin
        v:=-1; fi:=-fMaxFi;
       end;
    4: begin
        v:=-1; fi:=0;
       end;
    6: begin
        v:=-1; fi:=fMaxFi;
       end;
  end;

  u:=vector_u.create(expandingNode.ID,0,v,fi);  
  q_new:=Simulator.Integrate(expandingNode,u,fdti,2);
  if (q_new.x<5) or (q_new.y<5) or (q_new.x>795) or (q_new.y>595) then exit;

  pos:=PointLocation(q_new);
  if TCell2(FD[pos.i,pos.j]).occupied=true then
      begin
        u.Free; q_new.Free;
        exit;
      end;

  q_new.Depth:=expandingNode.Depth+1;
  Evaluate(q_new);
  if (expandingNode.Depth>10) then // od hloubky 10 vraci jen uzly s lepsim ohodnocenim
    if (expandingNode.Evaluation+5<q_new.Evaluation) then exit;

  if (TCell2(FD[pos.i,pos.j]).visited)=true then
    if vector_q(TCell2(FD[pos.i,pos.j]).node).Evaluation < q_new.Evaluation+1 then exit
     else TCell2(FD[pos.i,pos.j]).node:=q_new;  // lepsi ohodnoceni

  if (TCell2(FD[pos.i,pos.j]).visited)=false then
    begin
      TCell2(FD[pos.i,pos.j]).visited:=true;
      TCell2(FD[pos.i,pos.j]).node:=q_new;
    end;

  newID:= Nodes.Count;
  q_new.ID:=newID;
  AddNode(q_new);
    u.ToID:=q_new.ID;
  AddEdge(u);
  result:=q_new;
end;

function TGraphISSD.ISSD(aInit,aGoal:vector_q; aMaxDepth:integer):boolean;
var Open,Closed:TList;
    Found:boolean;
    ExpandingNode,newNode:vector_q;
    i,IndexO,IndexC:integer;
    pos: TPosition;
    d: double;
begin
     result:=false;
     CleanAll;
     pos:=PointLocation(aInit);
     if (TCell2(FD[pos.i,pos.j]).occupied)=true then
       begin
         MessageDlg('Startovn pozice se nachz v pekkov oblasti!',mtWarning,[mbOK],0);
         exit;
       end;
     if fDrawCells then DrawCells;
     AddNode(aInit);
     if IsEqualData(aInit,aGoal) then
      begin
        Solution.Add(aInit); result:=true; exit;
      end;
     Evaluate(aInit);
     Open:=TList.Create;
     Closed:=TList.Create;
     Found:=false;
     Open.Add(aInit);
     while (not found) and (Open.count>0) do
       begin
          if (fStop) then begin   // zastaveni vypoctu
            CleanAll;
            exit;
          end;
          ExpandingNode := vector_q(open.Items[Open.Count-1]);
          Closed.Add(ExpandingNode);
          Open.Delete(Open.Count-1);
          for i:=1 to RulesCount do
            begin
               newNode:= AddNodeToGraph(expandingNode,i);
               if newNode=nil then continue;
               if IsEqualData(newNode,aGoal) then
                begin
                  Solution.Add(NewNode);
                  RebuildSolution; found:=true; break;
                end;
               if newNode.Depth < aMaxDepth+1 then
                 begin
                   SortedList(open,newNode);
                   if fDrawTree then DrawTree(expandingNode,newNode);
                 end
               else begin
                 Open.Clear;
                 Closed.Clear;
                 exit;
               end;
            end;
       end;
     if found then result:=true;
     Open.Free;
     Closed.Free;
end;

function TGraphISSD.IsEqualData(aNode1, aNode2: TNode): boolean;
var d: double;
begin
  result:=false;
  if not(aNode1 is vector_q)or not(aNode2 is vector_q) then exit;
  d:=Distance.Distance(vector_q(aNode1),vector_q(aNode2));
  if (d<fEndD) then result:=true;
end;

procedure TGraphISSD.FindPath;
var q_near: vector_q;
         v: TVert;
       i,l: integer;
begin
  if (Solution.Count=0) then exit;
  SetLength(SolVertices,0);
  for i:=0 to Solution.Count-1 do
  begin
    v.x:=trunc(vector_q(Solution.Items[i]).x);
    v.y:=trunc(vector_q(Solution.Items[i]).y);
    SetLength(SolVertices,i+1);
    SolVertices[i]:=v;
  end;
end;

procedure TGraphISSD.DrawTree(q_near, q_new: vector_q);
var a,b: TVert;
    l: integer;
begin
  a.x:=trunc(q_near.x);
  a.y:=trunc(q_near.y);

  b.x:=trunc(q_new.x);
  b.y:=trunc(q_new.y);
  DrawTreeProc(a,b);
end;

procedure TGraphISSD.DrawCells;
var a,b: vertex;
    w,l,i: integer;
begin
  w:=trunc(fMaxX/fdt);
  l:=trunc(fMaxY/fdt);
  for i:=1 to w do
    begin
      a.x:=i*fdt;
      a.y:=0;
      b.x:=i*fdt;
      b.y:=l*fdt;
      DrawCellsProc(a,b);
    end;
  for i:=1 to l do
    begin
      a.x:=0;
      a.y:=i*fdt;
      b.x:=w*fdt;
      b.y:=i*fdt;
      DrawCellsProc(a,b);
    end;
end;

procedure TGraphISSD.ResetSol;
begin
  SetLength(SolVertices,0);
end;

end.
