//------------------------------------------------------------------------------------------
// The Rapidly-Exploring Random Tree
//------------------------------------------------------------------------------------------

unit GraphRRT;

interface

uses
  Graph, Math, Classes, Distance, CollisionDetection, Simulator, Polygon,
  Obstacles, Dialogs, SysUtils;

type

  TTree = array of TVert;
  TMode = record
    PG,PG1,D: double; // PG -pravdepodobnost ze nahodna konfigurace je cilova
  end;                // PG1 ci ze lezi ve vzdalenosti D od cile

  TGraphRRT = class (TGraph)        //RRT metoda
  private
    fMaxX: integer;   // celkova sirka plochy
    fMaxY: integer;   // celkova vyska plochy
    fIter: integer;   // pocet iteraci pri vypoctu
    fEndD: double;    // max.odchylka od cilove konf.(podm. ukonceni vypoctu)
    fRepeat: integer; // pocet opakovani vstupni akce
    fStop: boolean;   // zastaveni vypoctu
      // parametry modelu auta
    fCarL: double;		  // vzdalenost os
	  fMaxFi: double;		  // maximalni natoceni prednich kol
	  fCarWidth: double;	// sirka auta
  	fCarOver: double;		// presah pres osu vzadu a vpredu auta
	  fdt: integer;		  	// integracni krok
    fMode: TMode;
    fGoalMode: TMode;
    fDrawTree: boolean; //vykreslovani stromu behem vypoctu
    Simulator: TSimulator;
    Distance: TDistance;
    CollisionDetector: TCollisionDetection;
    SolVertices: TTree;  // pole vertex na cest k cli
    ObsList: TList;

    function GoalBiasedConf(goal:vector_q):vector_q;
    function RandConf:vector_q;
    function OptmInput(q_near,q_rand:vector_q):vector_u;
    function CollisionDetection(aq:vector_q):boolean;
    function OutBorder(aCar:TPolygon):boolean;
    function GetNearestConf(aConf:vector_q; fDist:TDistance): vector_q; //fdist-funktor vzdalenosti
  public
    constructor Create(aWidth,aLength:integer);
    destructor Destroy;override;
    procedure DrawTreeProc(a,b: TVert);virtual;abstract;
    procedure SetObstacles(aObstacles:TList);
    procedure SetMethod(aWidth,aHeight,aIter:integer; aEnd:double);
    procedure SetMode(aMode:TMode);
    procedure SetGoalMode(aMode:TMode);
    procedure DrawTree(q_near,q_new: vector_q);
    function BuildRRT(start,goal:vector_q):boolean;
    procedure FindPathRRT(aGoal:vector_q);
    property SolTree:TTree read SolVertices;
    property Collision:TCollisionDetection read CollisionDetector;
    property bDrawTree:boolean read fDrawTree write fDrawTree;
    property dt:integer read fdt write fdt;
    property NormalMode:TMode read fMode write fMode;
    property GoalMode:TMode read fGoalMode write fGoalMode;
    property EndD:double read fEndD write fEndD;
    property Iter:integer read fIter write fIter;
    property MaxFi:double read fMaxFi write fMaxFi;
    property Rep:integer read fRepeat write fRepeat;
    property Stop:boolean read fStop write fStop;
    property xSimulator:TSimulator read Simulator write Simulator;
    property Max_X:integer read fMaxX;
    property Max_Y:integer read fMaxY;
  end;

implementation

{ TGraphRRT }

constructor TGraphRRT.Create(aWidth,aLength:integer);
begin
  inherited create;
  fMaxX:=aWidth;
  fMaxY:=aLength;
  fIter:=1000;
  fCarL:=15;
  fMaxFi:=35*(Pi/180);
  fCarWidth:=23;
  fCarOver:=8;
  fdt:=8;       // integracni krok
  fEndD:=8;     // cilova tolerance
  fRepeat:=10;  // opakovani akce
  fStop:=false;
  fDrawTree:=true;
  Distance:=TDistance.Create;
  Distance.SetA(fCarL*fCarL);
  Simulator:=TSimulator.Create;
  Simulator.SetLength(fCarL);
  CollisionDetector:=TCollisionDetection.Create;
  fMode.PG:=0.1; fMode.PG1:=0.5; fMode.D:=300;
  fGoalMode.PG:=0.5; fGoalMode.PG1:=1; fGoalMode.D:=50;
end;

destructor TGraphRRT.Destroy;
begin
  Distance.Free;
  Simulator.Free;
  CollisionDetector.Free;
  inherited;
end;


function TGraphRRT.GetNearestConf(aConf:vector_q; fDist:TDistance): vector_q;
var minDist,d: double;
  nearestConf: vector_q;
  i:integer;
begin
  minDist:=maxdouble;
  d:=0;
  for i:=0 to Nodes.Count-1 do
    begin
      d:=fdist.Distance(aConf,vector_q(Nodes.Items[i]));
      if (d<minDist) then
        begin
          minDist:=d;
          nearestConf:=Nodes.Items[i];
        end;
    end;
  result:=nearestConf;
end;


  // Vraci nahodnou konfiguraci se spadem k cili
function TGraphRRT.GoalBiasedConf(goal: vector_q): vector_q;
var x: double;
   qr,q_rand: vector_q;
begin
  randomize;
  x:=random;  //rozsah[0..1]
  if (x<fMode.PG) then
    begin
      result:=goal;  //vrati cilovou konf.
      exit;
    end;
  if (x<fMode.PG1) then
    begin
      qr:=RandConf;
      while (Distance.Distance(qr,goal)>fMode.D) do
        qr:=RandConf;
      result:=qr;
      exit;
    end;
 result:=RandConf;
end;

  // Vraci nahodnou konfiguraci
function TGraphRRT.RandConf: vector_q;
var q_rand: vector_q;
  x,y,theta: double;
begin
  randomize;
  x:=random(fMaxX);
  y:=random(fMaxY);
  theta:=random*2*pi;   
  q_rand:=vector_q.create(x,y,theta); //vytvori novou instanci s daty
  result:=q_rand;
end;

  //optimalni vstup k ziskani qnew,ktery je nejblize z qnear k qrand    
function TGraphRRT.OptmInput(q_near, q_rand: vector_q): vector_u;
var q_new,q_opt,q1,q2: vector_q;
        u_opt,u,u1,u2: vector_u;
        i,j,s: integer;
        d,minD,d1,d2,fi: double;
        afi: array[0..2] of double;
        v: array[1..2] of integer;
        k1,k2: boolean;
begin
  v[1]:=-1; v[2]:=1;
  minD:=maxdouble;
    //derivace metriky podle dFi a rovno nule :(distance(qn,qr))^2  ->
  afi[0]:=arctan(fCarL*(q_near.theta-q_rand.theta)/ ((q_near.x-q_rand.x)*cos(q_near.theta)+(q_near.y-q_rand.y)*sin(q_near.theta)));
  afi[1]:=arctan(fCarL*(q_near.theta-q_rand.theta+2*PI)/ ((q_near.x-q_rand.x)*cos(q_near.theta)+(q_near.y-q_rand.y)*sin(q_near.theta)));
  afi[2]:=arctan(fCarL*(q_near.theta-q_rand.theta-2*PI)/ ((q_near.x-q_rand.x)*cos(q_near.theta)+(q_near.y-q_rand.y)*sin(q_near.theta)));
    //porovnani optm.vstupu (opt.natoceni kol)
  for i:=1 to 2 do
    for j:=0 to 2 do
      begin
        s:=v[i];
        fi:=afi[j];
        if (fi<-fMaxFi) then fi:=-fMaxFi
        else
          if (fi>fMaxFi) then fi:=fMaxFi;
        u:=vector_u.create(0,0,s,fi);
        q_new:=Simulator.Integrate(q_near,u,fdt,1);
        d:=Distance.Distance(q_rand,q_new);
        if (d<minD) then
          begin
            minD:=d;
            u_opt:=u;       // opt.akce (ukazatel na u)
            q_opt:=q_new;   // opt.konfigurace
          end
        else
          begin
            u.Free; q_new.Free;
          end;
      end;
  k1:=false;
  k2:=false;
  d1:=0;
  d2:=0;
        //vypocet noveho natoceni kol v pripade kolize
  if (CollisionDetection(q_opt)) then
        begin
          u1:=vector_u.create(q_near.ID,q_rand.ID,u_opt.s,u_opt.fi);
          u2:=vector_u.create(q_near.ID,q_rand.ID,u_opt.s,u_opt.fi);
          u1.fi:=u1.fi+Pi/20;   //+- maleho uhlu k optim.Fi
          u2.fi:=u2.fi-Pi/20;
          q1:=Simulator.Integrate(q_near,u1,fdt,1);
          q2:=Simulator.Integrate(q_near,u2,fdt,1);
          k1:=true;
          k2:=true;
        end;

    while k1 do     //pricitani maleho uhlu dokud dochazi ke kolizi
        begin
          k1:=false;
          d1:=Distance.Distance(q1,q_rand);
          if (CollisionDetection(q1)) then
            begin
              u1.fi:=u1.fi+Pi/20;
              q1:=Simulator.Integrate(q_near,u1,fdt,1);
              k1:=true;
            end;
          if (abs(u1.fi)>fMaxFi) then
            begin     //nelze se vyhnout prekazce zvetsenim uhlu Fi
              d1:=10000;
              break;
            end;
        end;

      while k2 do     //odecitani maleho uhlu dokud dochazi ke kolizi
        begin
          k2:=false;
          d2:=Distance.Distance(q2,q_rand);
          if (CollisionDetection(q2)) then
            begin
              u2.fi:=u2.fi-Pi/20;
              q2:=Simulator.Integrate(q_near,u2,fdt,1);
              k2:=true;
            end;
          if (abs(u2.fi)>fMaxFi) then
            begin     //nelze se vyhnout prekazce zvetsenim uhlu Fi
              d2:=10000;
              break;
            end;
        end;

      if (d1=10000) and (d2=10000) then
        u_opt.s:=0   //nelze se vyhnout prekazce

      else
        if (d1<d2) then
          u_opt:=u1
        else if (d1>d2) then
          u_opt:=u2;
  result:=u_opt;
end;

  // Detekce Kolizi mezi autem a prekazkami
function TGraphRRT.CollisionDetection(aq: vector_q): boolean;
var Car: TPolygon;
 v: array[0..3] of vertex;
 vertex: TVertex;
 i: integer;
 hlp: TPolygon;
begin
  result:=false;

  Car:=TPolygon.Create;
    //rohove body auta
  v[0].x:=-fCarOver;
  v[0].y:=fCarWidth/2;

  v[1].x:=fCarL+fCarOver;
  v[1].y:=fCarWidth/2;

  v[2].x:=fCarL+fCarOver;
  v[2].y:=-fCarWidth/2;

  v[3].x:=-fCarOver;
  v[3].y:=-fCarWidth/2;

  for i:=0 to 3 do
    begin
      vertex:=TVertex.create;
      vertex.x:=v[i].x;
      vertex.y:=v[i].y;
      vertex.Rotate(-aq.theta);
      vertex.Move(aq.x,aq.y);
      Car.AddVertex(vertex);
    end;

  if (OutBorder(Car)) then
    begin
      result:=true;
      exit;
    end;
  if (ObsList=nil) then exit;

  for i:=0 to ObsList.Count-1 do
   begin
    hlp:=TPolygon(ObsList.Items[i]);
    if (CollisionDetector.CollisionDetection(Car,hlp)) then
      begin
        result:=true;
        exit;
      end;
   end;
  Car.Free;
end;

  // Vraci true,pokud auto vyjelo mimo pracovni plochu
function TGraphRRT.OutBorder(aCar: TPolygon): boolean;
var i: integer;
begin
  result:=false;
  for i:=0 to aCar.VertexCount-1 do
    if (TVertex(aCar.Vertices.Items[i]).x>fMaxX) or (TVertex(aCar.Vertices.Items[i]).x<0) or
       (TVertex(aCar.Vertices.Items[i]).y>fMaxY) or (TVertex(aCar.Vertices.Items[i]).y<0) then
       result:=true;
end;


procedure TGraphRRT.SetObstacles(aObstacles: TList);
begin
  ObsList:=aObstacles;
end;


  // RRT metoda (vraci false kdyz vyhledavani trva dele nez 8s)
function TGraphRRT.BuildRRT(start, goal: vector_q):boolean;
var i,j: integer;
  q_rand,q_near,q_new: vector_q;
  u_opt: vector_u;
  newID: integer;
  newEdge: vector_u;
  d,dmin: double;
  clean: boolean;
begin
 result:=false;
 CleanAll;
 AddNode(start);
 d:=maxdouble;
  for i:=0 to fIter-1 do
    begin
      if (fStop) then begin   // zastaveni vypoctu
          CleanAll;
          result:=true;
          exit;
        end;
      q_rand:= GoalBiasedConf(goal);                // nahodna konf.
      q_near:= GetNearestConf(q_rand,Distance);     // nejblizsi konf.ve stromu ke q_rand
      u_opt:= OptmInput(q_near,q_rand);             // opt.vstup (akce robota)
      if (u_opt.s=0) then continue;                 // pri 0rychlosti do stromu nic nepridavej
        q_new:=Simulator.Integrate(q_near,u_opt,fdt,1); // nova konf.
      newID:=getMaxNodeID+1;
      q_new.ID:=newID;
      AddNode(q_new);                             // pridani uzlu do stromu
        u_opt.FRomID:=q_near.ID; u_opt.ToID:=q_new.ID;
      AddEdge(u_opt);                             // pridani hrany
      if fDrawTree then DrawTree(q_near,q_new);   // kresleni stromu
      d:=Distance.Distance(q_new,goal);
      if (d<fEndD) then begin
        result:=true;
        break;
      end;
      if (d<fGoalMode.D) then SetMode(fGoalMode);
          {*/ Pokus o opakovani stejne opt.akce /*}
      dmin:=maxdouble;
      for j:=0 to fRepeat do
      begin
        q_near:=q_new;
        q_new:=Simulator.Integrate(q_new,u_opt,fdt,1); // nova konf.
        d:=Distance.Distance(q_new,q_rand);
        if (d<dmin) then dmin:=d
          else break;     // vzdaluje se od q_rand
        if CollisionDetection(q_new) then break;    // kolizni konfigurace
        newID:=getMaxNodeID+1;
        q_new.ID:=newID;
        AddNode(q_new);                             // pridani uzlu do stromu
        newEdge:=vector_u.create(q_near.ID,q_new.ID,u_opt.s,u_opt.fi);
        AddEdge(newEdge);                           // pridani hrany
        if fDrawTree then DrawTree(q_near,q_new);   // kresleni stromu
        d:=Distance.Distance(q_new,goal);
        if (d<fEndD) then begin
          result:=true;
          break;
        end;
        if (d<fGoalMode.D) then SetMode(fGoalMode);
      end;
    end;
end;


procedure TGraphRRT.FindPathRRT(aGoal:vector_q);
var q_near: vector_q;
         v: TVert;
       i,l: integer;
begin
  SetLength(SolVertices,0);
  q_near:=GetNearestConf(aGoal,Distance);
  if (Distance.Distance(aGoal,q_near)<fEndD) then
    begin
      Solution.Add(q_near);
      if (RebuildSolution) then
        begin
          for i:=0 to Solution.Count-1 do
            begin
              v.x:=trunc(vector_q(Solution.Items[i]).x);
              v.y:=trunc(vector_q(Solution.Items[i]).y);
              SetLength(SolVertices,i+1);
              SolVertices[i]:=v;
            end;
        end
      else
        MessageDlg('een nebylo nalezeno.',mtError,[mbOk],0);
    end
end;


procedure TGraphRRT.SetMethod(aWidth, aHeight, aIter: integer; aEnd:double);
begin
  fMaxX:=aWidth;
  fMaxY:=aHeight;
  fIter:=aIter;
  fEndD:=aEnd;
end;


procedure TGraphRRT.DrawTree(q_near, q_new: vector_q);
var a,b: TVert;
    l: integer;
begin
  a.x:=trunc(q_near.x);
  a.y:=trunc(q_near.y);

  b.x:=trunc(q_new.x);
  b.y:=trunc(q_new.y);
  DrawTreeProc(a,b);
end;


procedure TGraphRRT.SetMode(aMode:TMode);
begin
  fMode.PG:=aMode.PG;
  fMode.PG1:=aMode.PG1;
  fMode.D:=aMode.D;
end;

procedure TGraphRRT.SetGoalMode(aMode:TMode);
begin
  fGoalMode.PG:=aMode.PG;
  fGoalMode.PG1:=aMode.PG1;
  fGoalMode.D:=aMode.D;
end;



end.

