unit Simulator;     //? prechod.stav.rce

interface

uses
  Graph, SysUtils;

type                            // data
  vector_q = class (TNode)      // konfigurace robotu
    private
      fx,fy,ftheta: double;
    public
      property x:double read fx write fx;
      property y:double read fy write fy;
      property theta:double read ftheta write ftheta;
      constructor create(aID:TID; ax,ay,atheta:double);overload;
      constructor create(ax,ay,atheta:double);overload;
  end;

  vector_u = class (TEdge)     //ridici vektor robotu
    private
      Fs,Ffi: double;
    public
      property s:double read fs write fs;
      property fi:double read ffi write ffi;
      constructor create(aFromID,aToID:TID;s,fi:double);
  end;

  TSimulator = Class (TObject)
  private
    fL: double;
    fMethod: boolean; // int.metoda (true=E, false=R-K)
    function Equation1(aq:vector_q; au:vector_u):vector_q;
    function Equation2(aq:vector_q; au:vector_u):vector_q;
    function EulerMethod(aq:vector_q; au: vector_u; dt:integer; i:integer):vector_q;
    function RungeKuttaMethod(aq:vector_q; au: vector_u; dt:integer; i:integer):vector_q;
  public
    constructor Create;
    property Method:boolean read fMethod write fMethod;
    procedure SetLength(aL:double);
    function Integrate(aq:vector_q; au:vector_u; dt:integer; i:integer):vector_q;
  end;

implementation

constructor vector_q.create(aID:TID; ax,ay,atheta:double);
begin
  inherited create(aID);
  fx:=ax;
  fy:=ay;
  ftheta:=atheta;
end;

constructor vector_q.create(ax,ay,atheta:double);
begin
  fx:=ax;
  fy:=ay;
  ftheta:=atheta;
end;

constructor vector_u.create(aFromID,aToID:TID;s,fi:double);
begin
  inherited create(aFromID,aToID);
  Fs:=s;
  Ffi:=fi;
end;

constructor TSimulator.Create;
begin
  inherited create;
  fMethod:=false;
  fL:=0;
end;

procedure TSimulator.SetLength(aL:double);
begin
  fL:=aL;
end;

  // Soustava diferencialnich rovnic popisujicich pohyb robotu (pro RRT)
function TSimulator.Equation1(aq:vector_q; au:vector_u):vector_q;
var qn: vector_q;
  dx,dy,dtheta: double;
begin
  result:=nil;                        // nahon na predni kola
  dx:=au.s*cos(aq.theta)*cos(au.fi);  // x
  dy:=au.s*sin(aq.theta)*cos(au.fi);  // y
  dtheta:=(au.s/fL)*sin(au.fi);       // celkove natoceni
  qn:=vector_q.create(dx,dy,dtheta);
  result:=qn;
end;

  // Soustava diferencialnich rovnic popisujicich pohyb robotu (pro IGPPR a ISSD)
function TSimulator.Equation2(aq:vector_q; au:vector_u):vector_q;
var qn: vector_q;
  dx,dy,dtheta: double;
begin
  result:=nil;                        // nahon na predni kola
  dx:=au.s*cos(aq.theta)*cos(au.fi);  // x
    dx:=dx-0.15;
  dy:=au.s*sin(aq.theta)*cos(au.fi);  // y
    if au.fi>0 then dy:=dy+0.15
      else if au.fi<0 then dy:=dy-0.15;
  dtheta:=(au.s/fL)*sin(au.fi);       // celkove natoceni
  qn:=vector_q.create(dx,dy,dtheta);
  result:=qn;
end;

function TSimulator.Integrate(aq:vector_q; au:vector_u; dt:integer; i:integer):vector_q;
begin
  if fMethod then result:=EulerMethod(aq,au,dt,i) else
    result:=RungeKuttaMethod(aq,au,dt,i);
end;

function TSimulator.EulerMethod(aq:vector_q; au: vector_u; dt:integer; i:integer):vector_q;
var qn: vector_q;
   str: string;
begin         //dq -rychlost
  if i=1 then qn:=Equation1(aq,au) else
    qn:=Equation2(aq,au);
  qn.x:= aq.x + qn.x*dt;
  qn.y:= aq.y + qn.y*dt;
  qn.theta:= aq.theta + qn.theta*dt;
  result:=qn;
end;

function TSimulator.RungeKuttaMethod(aq:vector_q; au: vector_u; dt:integer; i:integer):vector_q;
var w1,w2,w3,w4,qtmp,qn:vector_q;
begin
  if i=1 then w1:=Equation1(aq,au) else
    w1:=Equation2(aq,au);

  qtmp:=vector_q.create(0);
  qtmp.x:=aq.x+0.5*dt*w1.x;
  qtmp.y:=aq.y+0.5*dt*w1.y;
  qtmp.theta:=aq.theta+0.5*dt*w1.theta;
  if i=1 then w2:=Equation1(qtmp,au) else
    w2:=Equation2(qtmp,au);

  qtmp.x:=aq.x+0.5*dt*w2.x;
  qtmp.y:=aq.y+0.5*dt*w2.y;
  qtmp.theta:=aq.theta+0.5*dt*w2.theta;
  if i=1 then w3:=Equation1(qtmp,au) else
    w3:=Equation2(qtmp,au);

  qtmp.x:=aq.x+dt*w3.x;
  qtmp.y:=aq.y+dt*w3.y;
  qtmp.theta:=aq.theta+dt*w3.theta;
  if i=1 then w4:=Equation1(qtmp,au) else
    w4:=Equation2(qtmp,au);

  qtmp.x:=aq.x+dt/6*(w1.x+2*w2.x+2*w3.x+w4.x);
  qtmp.y:=aq.y+dt/6*(w1.y+2*w2.y+2*w3.y+w4.y);
  qtmp.theta:=aq.theta+dt/6*(w1.theta+2*w2.theta+2*w3.theta+w4.theta);

  result:=qtmp;
end;



end.


